# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
from pathlib import Path
import nevergrad as ng
from nevergrad.parametrization import FolderFunction


if __name__ == "__main__":
    folder = Path(__file__).parents[1] / "parametrization" / "examples"
    func = FolderFunction(folder, [sys.executable, "examples/script.py"], clean_copy=True)
    instrumentation = ng.p.Instrumentation(
        value1=ng.p.Scalar(), value2=12, string=ng.p.Choice(["plop", "blublu", "plouf"])
    )
    opt = ng.optimizers.registry["OnePlusOne"](instrumentation, budget=4)
    opt.minimize(func)
    ng.families.ParametrizedOnePlusOne()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Based on a discussion at Dagstuhl's seminar on Computational Intelligence in Games with:
# - Dan Ashlock
# - Chiara Sironi
# - Guenter Rudolph
# - Jialin Liu

import numpy as np
from nevergrad.parametrization import parameter as p
from ..base import ExperimentFunction
from scipy.ndimage import gaussian_filter


class TO(ExperimentFunction):
    def __init__(self, n: int = 50) -> None:
        super().__init__(
            self._simulate_to, p.Array(shape=(n, n), lower=-1.0, upper=1.0).set_name(f"array{n}x{n}")
        )
        self.n = n
        self.idx = self.parametrization.random_state.randint(50000)

    def _simulate_to(self, x: np.ndarray) -> float:
        x = x.reshape(self.n, self.n)
        idx = self.idx
        n = self.n
        xa = idx % 3
        size = n * n
        sqrtsize = n
        xb = 2 - xa
        if (idx // 12) % 2 > 0:
            xs = 1.5 * (
                np.array(
                    [
                        float(np.cos(self.idx * 0.01 + xa * i + xb * j) < 0.0)
                        for i in range(n)
                        for j in range(n)
                    ]
                ).reshape(n, n)
                - 0.5
            )
        else:
            xs = 1.5 * (
                np.array(
                    [float((self.idx * 0.01 + xa * i + xb * j) > 1.6 * n) for i in range(n) for j in range(n)]
                ).reshape(n, n)
                - 0.5
            )
        if (idx // 3) % 2 > 0:
            xs = np.transpose(xs)
        if (idx // 6) % 2 > 0:
            xs = -xs
        return (
            5.0 * np.sum(np.abs(x - xs) > 0.3) / size
            + 3.0 * np.linalg.norm(x - gaussian_filter(x, sigma=3)) / sqrtsize
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import nevergrad.common.typing as tp
from nevergrad.parametrization import discretization

from . import utils


class Mutator:
    """Class defining mutations, and holding a random state used for random generation."""

    def __init__(self, random_state: np.random.RandomState) -> None:
        self.random_state = random_state

    def significantly_mutate(self, v: float, arity: int):
        """Randomly drawn a normal value, and redraw until it's different after discretization by the quantiles
        1/arity, 2/arity, ..., (arity-1)/arity.
        """
        w = v
        while discretization.threshold_discretization([w], arity) == discretization.threshold_discretization(
            [v], arity
        ):
            w = self.random_state.normal(0.0, 1.0)
        return w

    def doerr_discrete_mutation(self, parent: tp.ArrayLike, arity: int = 2) -> tp.ArrayLike:
        """Mutation as in the fast 1+1-ES, Doerr et al. The exponent is 1.5."""
        dimension = len(parent)
        if dimension < 5:
            return self.discrete_mutation(parent)
        return self.doubledoerr_discrete_mutation(parent, max_ratio=0.5, arity=arity)

    def doubledoerr_discrete_mutation(
        self, parent: tp.ArrayLike, max_ratio: float = 1.0, arity: int = 2
    ) -> tp.ArrayLike:
        """Doerr's recommendation above can mutate up to half variables
        in average.
        In our high-arity context, we might need more than that.

        Parameters
        ----------
        parent: array-like
            the point to mutate
        max_ratio: float (between 0 and 1)
            the maximum mutation ratio (careful: this is not an exact ratio)
        """
        assert 0 <= max_ratio <= 1
        dimension = len(parent)
        max_mutations = max(2, int(max_ratio * dimension))
        p = 1.0 / np.arange(1, max_mutations) ** 1.5
        p /= np.sum(p)
        u = self.random_state.choice(np.arange(1, max_mutations), p=p)
        return self.portfolio_discrete_mutation(parent, intensity=u, arity=arity)

    def rls_mutation(self, parent: tp.ArrayLike, arity: int = 2) -> tp.ArrayLike:
        """Good old one-variable mutation.

        Parameters
        ----------
        parent: array-like
            the point to mutate
        arity: int
            the number of possible distinct values
        """
        dimension = len(parent)
        if dimension == 1:  # corner case.
            return self.random_state.normal(0.0, 1.0, size=1)  # type: ignore
        out = np.array(parent, copy=True)
        ind = self.random_state.randint(dimension)
        out[ind] = self.significantly_mutate(out[ind], arity)
        return out

    def portfolio_discrete_mutation(
        self, parent: tp.ArrayLike, intensity: tp.Optional[int] = None, arity: int = 2
    ) -> tp.ArrayLike:
        """Mutation discussed in
        https://arxiv.org/pdf/1606.05551v1.pdf
        We mutate a randomly drawn number of variables on average.
        The mutation is the same for all variables - coordinatewise mutation will be different from this point of view and will make it possible
        to do anisotropic mutations.
        """
        dimension = len(parent)
        if intensity is None:
            intensity = 1 if dimension == 1 else int(self.random_state.randint(1, dimension))
        if dimension == 1:  # corner case.
            return self.random_state.normal(0.0, 1.0, size=1)  # type: ignore
        boolean_vector = np.ones(dimension, dtype=bool)
        while all(boolean_vector) and dimension != 1:
            boolean_vector = self.random_state.rand(dimension) > float(intensity) / dimension
        return [s if b else self.significantly_mutate(s, arity) for (b, s) in zip(boolean_vector, parent)]

    def coordinatewise_mutation(
        self,
        parent: tp.ArrayLike,
        velocity: tp.ArrayLike,
        boolean_vector: tp.ArrayLike,
        arity: int,
    ) -> tp.ArrayLike:
        """This is the anisotropic counterpart of the classical 1+1 mutations in discrete domains
        with tunable intensity: it is useful for anisotropic adaptivity."""
        dimension = len(parent)
        boolean_vector = np.zeros(dimension, dtype=bool)
        while not any(boolean_vector):
            boolean_vector = self.random_state.rand(dimension) < (1.0 / dimension)
        discrete_data = discretization.threshold_discretization(parent, arity=arity)
        discrete_data = np.where(  # type: ignore
            boolean_vector,
            discrete_data + self.random_state.choice([-1.0, 1.0], size=dimension) * velocity,
            discrete_data,
        )
        return discretization.inverse_threshold_discretization(discrete_data)

    def discrete_mutation(self, parent: tp.ArrayLike, arity: int = 2) -> tp.ArrayLike:
        """This is the most classical discrete 1+1 mutation of the evolution literature."""
        dimension = len(parent)
        boolean_vector = np.ones(dimension, dtype=bool)
        while all(boolean_vector):
            boolean_vector = self.random_state.rand(dimension) > (1.0 / dimension)
        return [s if b else self.significantly_mutate(s, arity) for (b, s) in zip(boolean_vector, parent)]

    def crossover(
        self, parent: tp.ArrayLike, donor: tp.ArrayLike, rotation: bool = False, crossover_type: str = "none"
    ) -> tp.ArrayLike:
        if rotation:
            dim = len(parent)
            k = self.random_state.randint(1, dim)
            mix = [self.random_state.choice([donor[(i + k) % dim], parent[i]]) for i in range(len(parent))]
        else:
            if crossover_type == "rand":
                crossover_type = str(self.random_state.choice(["max", "min", "onepoint", "twopoint"]))
            if crossover_type == "max":
                mix = [min([d, p]) for (p, d) in zip(parent, donor)]
            elif crossover_type == "min":
                mix = [max([d, p]) for (p, d) in zip(parent, donor)]
            elif crossover_type == "onepoint" and len(parent) > 4:
                sig = self.random_state.choice([-1.0, 1.0])
                idx = self.random_state.randint(len(parent) - 1) + 0.5
                mix = [(d if (i - idx) * sig < 0 else p) for i, (p, d) in enumerate(zip(parent, donor))]
            elif crossover_type == "twopoint" and len(parent) > 6:
                sig = self.random_state.choice([-1.0, 1.0])
                idx = self.random_state.randint(len(parent) - 1) + 0.5
                idx2 = self.random_state.randint(len(parent) - 1) + 0.5
                while idx == idx2:
                    idx2 = self.random_state.randint(len(parent) - 1) + 0.5
                mix = [
                    (d if (i - idx) * (i - idx2) * sig < 0 else p)
                    for i, (p, d) in enumerate(zip(parent, donor))
                ]
            else:
                mix = [self.random_state.choice([d, p]) for (p, d) in zip(parent, donor)]
        return self.discrete_mutation(mix)

    def get_roulette(self, archive: utils.Archive[utils.MultiValue], num: tp.Optional[int] = None) -> tp.Any:
        """Apply a roulette tournament selection."""
        if num is None:
            num = int(0.999 + np.sqrt(len(archive)))
        # the following sort makes the line deterministic, and function seedable, at the cost of complexity!
        my_keys = sorted(archive.bytesdict.keys())
        my_keys_indices = self.random_state.choice(len(my_keys), size=min(num, len(my_keys)), replace=False)
        my_keys = [my_keys[i] for i in my_keys_indices]
        # best pessimistic value in a random set of keys
        return np.frombuffer(min(my_keys, key=lambda x: archive.bytesdict[x].pessimistic_confidence_bound))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pickle
import numpy as np
import pytest
import nevergrad as ng
from nevergrad.common import testing
from nevergrad.functions import ArtificialFunction
import nevergrad.common.typing as tp
from . import base
from . import helpers


def _arg_return(*args: tp.Any, **kwargs: tp.Any) -> float:
    # for the sake of tests just do some really stupid return
    return args, kwargs  # type: ignore


def test_experiment_function() -> None:
    param = ng.p.Instrumentation(
        ng.p.Choice([1, 12]),
        "constant",
        ng.p.Array(shape=(2, 2)),
        constkwarg="blublu",
        plop=ng.p.Choice([3, 4]),
    )
    with pytest.raises(RuntimeError):
        base.ExperimentFunction(_arg_return, param)
    param.set_name("myparam")
    ifunc = base.ExperimentFunction(_arg_return, param)
    np.testing.assert_equal(ifunc.dimension, 8)
    data = [-100.0, 100, 1, 2, 3, 4, 100, -100]
    args0, kwargs0 = ifunc.parametrization.spawn_child().set_standardized_data(data).value
    output: tp.Any = ifunc(*args0, **kwargs0)
    args: tp.Any = output[0]
    kwargs: tp.Any = output[1]
    testing.printed_assert_equal(args, [12, "constant", [[1, 2], [3, 4]]])
    testing.printed_assert_equal(kwargs, {"constkwarg": "blublu", "plop": 3})
    testing.printed_assert_equal(
        ifunc.descriptors,
        {
            "dimension": 8,
            "name": "_arg_return",
            "function_class": "ExperimentFunction",
            "parametrization": "myparam",
        },
    )


def test_array_experiment_function() -> None:
    iarrayfuncs = [
        base.ArrayExperimentFunction(
            sum,
            ng.p.Array(shape=(10,)).set_bounds(-0.5, 6.0).set_name(""),
            symmetry=s,
        )
        for s in [247, 111, 111]
    ]
    np.testing.assert_equal(iarrayfuncs[0].dimension, 10)
    assert iarrayfuncs[0](np.zeros(10)) == iarrayfuncs[0].copy()(np.zeros(10))
    assert iarrayfuncs[0](np.zeros(10)) == 16.5
    assert iarrayfuncs[1](np.zeros(10)) == 22.0
    assert iarrayfuncs[0](np.ones(10)) != iarrayfuncs[1](np.ones(10))
    assert iarrayfuncs[2](np.ones(10)) == iarrayfuncs[1](np.ones(10))


def test_instrumented_function_kwarg_order() -> None:
    ifunc = base.ExperimentFunction(
        _arg_return,
        ng.p.Instrumentation(
            kw4=ng.p.Choice([1, 0]),
            kw2="constant",
            kw3=ng.p.Array(shape=(2, 2)),
            kw1=ng.p.Scalar(2.0).set_mutation(sigma=2.0),
        ).set_name("test"),
    )
    np.testing.assert_equal(ifunc.dimension, 7)
    data = np.array([-1, 1, 2, 3, 4, 100, -100])
    args0, kwargs0 = ifunc.parametrization.spawn_child().set_standardized_data(data).value
    # this is very stupid and should be removed when Parameter is in use
    kwargs: tp.Any = ifunc(*args0, **kwargs0)[1]  # type: ignore
    testing.printed_assert_equal(kwargs, {"kw1": 0, "kw2": "constant", "kw3": [[1, 2], [3, 4]], "kw4": 1})


class _Callable:
    def __call__(self, x: float, y: float = 0) -> float:
        return abs(x + y)


def test_callable_parametrization() -> None:
    ifunc = base.ExperimentFunction(lambda x: x**2, ng.p.Scalar(2).set_mutation(2).set_name(""))  # type: ignore
    np.testing.assert_equal(ifunc.descriptors["name"], "<lambda>")
    ifunc = base.ExperimentFunction(_Callable(), ng.p.Scalar(2).set_mutation(sigma=2).set_name(""))
    np.testing.assert_equal(ifunc.descriptors["name"], "_Callable")
    # test automatic filling
    assert len(ifunc._auto_init) == 2


def test_packed_function() -> None:
    ifunc = base.ExperimentFunction(_Callable(), ng.p.Scalar(1).set_name(""))
    with pytest.raises(AssertionError):
        base.MultiExperiment([ifunc, ifunc], [100, 100])
    pfunc = base.MultiExperiment([ifunc, ifunc.copy()], [100, 100])
    np.testing.assert_equal(pfunc.descriptors["name"], "_Callable,_Callable")
    np.testing.assert_array_equal(pfunc(-3), [3, 3])


def test_deterministic_data_setter() -> None:
    instru = ng.p.Instrumentation(ng.p.Choice([0, 1, 2, 3]), y=ng.p.Choice([0, 1, 2, 3])).set_name("")
    ifunc = base.ExperimentFunction(_Callable(), instru)
    data = [0.01, 0, 0, 0, 0.01, 0, 0, 0]
    for _ in range(20):
        child = ifunc.parametrization.spawn_child()
        with ng.p.helpers.deterministic_sampling(child):
            args, kwargs = child.set_standardized_data(data).value
        testing.printed_assert_equal(args, [0])
        testing.printed_assert_equal(kwargs, {"y": 0})
    arg_sum, kwarg_sum = 0, 0
    for _ in range(24):
        args, kwargs = ifunc.parametrization.spawn_child().set_standardized_data(data).value
        arg_sum += args[0]
        kwarg_sum += kwargs["y"]
    assert arg_sum != 0
    assert kwarg_sum != 0


@testing.parametrized(
    floats=((ng.p.Scalar(), ng.p.Scalar(init=12.0)), True, False),
    array_int=((ng.p.Scalar(), ng.p.Array(shape=(1,)).set_integer_casting()), False, False),
    softmax_noisy=((ng.p.Choice(["blue", "red"]), ng.p.Array(shape=(1,))), True, True),
    softmax_deterministic=(
        (ng.p.Choice(["blue", "red"], deterministic=True), ng.p.Array(shape=(1,))),
        False,
        False,
    ),
    ordered_discrete=((ng.p.TransitionChoice([True, False]), ng.p.Array(shape=(1,))), False, False),
)
def test_parametrization_continuous_noisy(
    variables: tp.Tuple[ng.p.Parameter, ...], continuous: bool, noisy: bool
) -> None:
    instru = ng.p.Instrumentation(*variables)
    analysis = ng.p.helpers.analyze(instru)
    assert analysis.continuous == continuous
    assert analysis.deterministic != noisy


class ExampleFunction(base.ExperimentFunction):
    def __init__(self, dimension: int, number: int, default: int = 12):  # pylint: disable=unused-argument
        # unused argument is used to check that it is automatically added as descriptor
        super().__init__(self.oracle_call, ng.p.Array(shape=(dimension,)))

    def oracle_call(self, x: np.ndarray) -> float:
        return float(x[0])

    # pylint: disable=unused-argument
    def compute_pseudotime(self, input_parameter: tp.Any, loss: tp.Loss) -> float:
        assert isinstance(loss, (int, float))
        return 5 - loss


def test_function_descriptors_and_pickle() -> None:
    func = ExampleFunction(dimension=1, number=3)
    assert "default" in func.descriptors
    assert "self" not in func._auto_init
    out = pickle.dumps(func)
    func2 = pickle.loads(out).copy()
    assert func2.descriptors["number"] == 3


class ExampleFunctionAllDefault(base.ExperimentFunction):
    def __init__(self, dimension: int = 2, default: int = 12):  # pylint: disable=unused-argument
        # unused argument is used to check that it is automatically added as descriptor
        super().__init__(lambda x: 3.0, ng.p.Array(shape=(dimension,)))


def test_function_descriptors_all_default() -> None:
    func = ExampleFunctionAllDefault()
    assert func.descriptors["default"] == 12
    assert "self" not in func._auto_init
    with pytest.raises(TypeError):
        # make sure unexpected keyword still works
        ExampleFunctionAllDefault(blublu=12)  # type: ignore


def test_pareto_experiment() -> None:
    # Checking MOO in cross-validation.
    objective_functions: tp.List[tp.Any] = [
        ArtificialFunction("sphere", block_dimension=7),
        ArtificialFunction("sphere", block_dimension=7),
        ArtificialFunction("cigar", block_dimension=7),
    ]
    xps = helpers.SpecialEvaluationExperiment.create_crossvalidation_experiments(
        objective_functions, pareto_size=16
    )
    assert len(xps) == 15  # 3 xps, multiplied by 5 Pareto extractors
    param = xps[0].parametrization
    out = xps[0](*param.args, **param.kwargs)
    assert isinstance(out, np.ndarray) and out.size == 2
    param._losses = out  # hack for testing
    evaluation = xps[0].evaluation_function(param, param, param)
    assert isinstance(evaluation, float)


def test_easy_pareto_experiment() -> None:
    # Checking MOO in cross-validation.
    objective_functions: tp.List[tp.Any] = [
        ArtificialFunction("sphere", block_dimension=2, translation_factor=0.0) for _ in range(3)
    ]
    xps = helpers.SpecialEvaluationExperiment.create_crossvalidation_experiments(
        objective_functions, pareto_size=16
    )
    print(xps[0])
    optimizer = ng.optimizers.OnePlusOne(parametrization=2, budget=100)
    optimizer.minimize(xps[0], verbosity=2)
    assert xps[0].evaluation_function(optimizer.pareto_front()[0]) ** 2 < 0.001



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from nevergrad.optimization import multiobjective as mobj
from . import base


class SpecialEvaluationExperiment(base.ExperimentFunction):
    """Experiment which uses one experiment for the optimization,
    and another for the evaluation
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        experiment: base.ExperimentFunction,
        evaluation: base.ExperimentFunction,
        pareto_size: tp.Optional[int] = None,
        pareto_subset: str = "random",
        pareto_subset_tentatives: int = 30,
    ) -> None:
        self._experiment = experiment
        self._evaluation = evaluation
        self._pareto_size = pareto_size
        self._pareto_subset = pareto_subset
        self._pareto_subset_tentatives = pareto_subset_tentatives
        super().__init__(self._delegate_to_experiment, experiment.parametrization)
        self.add_descriptors(non_proxy_function=False)
        # remove multiobjective descriptors if singleobjective / no pareto subset
        if self._pareto_size is None:
            names = [name for name in self._descriptors if name.startswith("pareto_")]
            for name in names:
                del self._descriptors[name]

    def _delegate_to_experiment(self, *args: tp.Any, **kwargs: tp.Any) -> tp.Loss:
        return self._experiment(*args, **kwargs)

    def copy(self) -> "SpecialEvaluationExperiment":
        """Creates with new experiments and evaluations"""
        instance = super().copy()
        for name in [
            "_experiment",
            "_evaluation",
        ]:
            setattr(instance, name, getattr(self, name).copy())
        return instance

    def compute_pseudotime(  # pylint: disable=unused-argument
        self, input_parameter: tp.Any, loss: tp.Loss
    ) -> float:
        return self._experiment.compute_pseudotime(input_parameter, loss)

    def evaluation_function(self, *recommendations: p.Parameter) -> float:
        if self._pareto_size is not None and len(recommendations) > self._pareto_size:
            # select a subset
            hypervolume = mobj.HypervolumePareto(upper_bounds=self.multiobjective_upper_bounds)
            hypervolume.extend(recommendations)
            recommendations = tuple(
                hypervolume.pareto_front(
                    size=self._pareto_size,
                    subset=self._pareto_subset,
                    subset_tentatives=self._pareto_subset_tentatives,
                )
            )
        return min(self._evaluation.evaluation_function(recom) for recom in recommendations)

    @property
    def descriptors(self) -> tp.Dict[str, tp.Any]:
        """Description of the function parameterization, as a dict. This base class implementation provides function_class,
        noise_level, transform and dimension
        """
        desc = dict(self._descriptors)
        desc.update(self._experiment.descriptors)
        # TODO descriptors for the evaluation function?
        return desc

    @classmethod
    def create_crossvalidation_experiments(
        cls,
        experiments: tp.List[base.ExperimentFunction],
        training_only_experiments: tp.Sequence[base.ExperimentFunction] = (),
        pareto_size: int = 12,
        pareto_subset_methods: tp.Sequence[str] = (
            "random",
            "loss-covering",
            "EPS",
            "domain-covering",
            "hypervolume",
        ),
    ) -> tp.List["SpecialEvaluationExperiment"]:
        """Returns a list of MultiExperiment, corresponding to MOO cross-validation:
        Each experiments consist in optimizing all but one of the input ExperimentFunction's,
        and then considering that the score is the performance of the best solution in the
        approximate Pareto front for the excluded ExperimentFunction.

        Parameters
        ----------
        experiments: sequence of ExperimentFunction
            iterable of experiment functions, used for creating the crossvalidation.
        training_only_experiments: sequence of ExperimentFunction
            iterable of experiment functions, used only as training functions in the crossvalidation and never for test..
        pareto_size: int
            if provided, selects a subset of the full pareto front with the given maximum size
        subset: str
            method for selecting the subset (see optimizer.pareto_front)

        """
        funcs: tp.List["SpecialEvaluationExperiment"] = []
        if "PYTEST_NEVERGRAD" in os.environ:
            pareto_subset_methods = ("random",)  # override for speed
        for pareto_subset in pareto_subset_methods:
            params: tp.Dict[str, tp.Any] = dict(pareto_size=pareto_size, pareto_subset=pareto_subset)
            for eval_xp in experiments:
                trainxps = [xp for xp in experiments if xp != eval_xp]
                if training_only_experiments is not None:
                    trainxps += training_only_experiments
                if len(trainxps) == 1:  # singleobjective
                    experiment = trainxps[0]
                else:  # multiobjective
                    # uses origin as upper bound
                    param = eval_xp.parametrization
                    upper_bounds = [xp(*param.args, **param.kwargs) for xp in trainxps]
                    experiment = base.MultiExperiment(trainxps, upper_bounds)  # type: ignore
                funcs.append(cls(experiment, eval_xp, **params))
        return funcs



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from unittest import SkipTest
import warnings
import typing as tp
import pytest
import numpy as np
from nevergrad.common import testing
from . import core
from . import photonics


@testing.parametrized(
    bragg=("bragg", [2.93, 2.18, 2.35, 2.12, 45.77, 37.99, 143.34, 126.55]),
    morpho=("morpho", [280.36, 52.96, 208.16, 72.69, 89.92, 60.37, 226.69, 193.11]),
    chirped=("chirped", [170.18, 56.48, 82.04, 48.17, 45.77, 37.99, 143.34, 126.55]),
)
def test_photonics_bounding_methods(pb: str, expected: tp.List[float]) -> None:
    func = core.Photonics(pb, 8, bounding_method="tanh")
    np.random.seed(24)
    x = np.random.normal(0, 1, size=8)
    output = func.parametrization.spawn_child().set_standardized_data(x).value.ravel()
    np.testing.assert_almost_equal(output, expected, decimal=2)


@testing.parametrized(
    # bragg domain (n=60): [2,3]^30 x [30,180]^30
    bragg=("bragg", [2.5, 2.5, 2.5, 2.5, 105.0, 105.0, 105.0, 105.0]),
    # chirped domain (n=60): [30,170]^60
    chirped=("chirped", [105.0, 105.0, 105.0, 105.0, 105.0, 105.0, 105.0, 105.0]),
    # morpho domain (n=60): [0,300]^15 x [0,600]^15 x [30,600]^15 x [0,300]^15
    morpho=("morpho", [150.0, 150.0, 300.0, 300.0, 315.0, 315.0, 150.0, 150.0]),
)
def test_photonics_mean(pb: str, expected: tp.List[float]) -> None:
    func = core.Photonics(pb, 8)
    all_x = func.parametrization.value
    output = all_x.ravel()
    np.testing.assert_almost_equal(output, expected, decimal=2)


def test_morpho_bounding_method_constraints() -> None:
    func = core.Photonics("morpho", 60, bounding_method="arctan")
    x = np.random.normal(0, 5, size=60)  # std 5 to play with boundaries
    output1 = func.parametrization.spawn_child().set_standardized_data(x)
    output2 = output1.sample()
    for output in (output1, output2):
        assert np.all(output.value >= 0)
        assert np.all(output.value[[0, 3], :] <= 300)
        assert np.all(output.value[[1, 2], :] <= 600)
        assert np.all(output.value[2, :] >= 30)


def test_photonics_bragg_recombination() -> None:
    func = core.Photonics("bragg", 8)
    func.parametrization.random_state.seed(24)
    arrays = [func.parametrization.spawn_child() for _ in range(2)]
    arrays[0].value = [[2, 2, 2, 2], [35, 35, 35, 35]]
    arrays[1].value = [[3, 3, 3, 3], [45, 45, 45, 45]]
    arrays[0].recombine(arrays[1])
    expected = [[3, 2, 2, 2], [45, 35, 35, 35]]
    np.testing.assert_array_equal(arrays[0].value, expected)


def test_photonics_custom_mutation() -> None:
    if os.environ.get("CIRCLECI", False):
        raise SkipTest("Skipping in CI because way too slow on their machine (weird)")
    func = core.Photonics("morpho", 16, rolling=True)
    param = func.parametrization.spawn_child()
    for _ in range(10):
        param.mutate()
    # as tuple
    func = core.Photonics("morpho", 16, rolling=True, as_tuple=True)
    func.evaluation_function(func.parametrization)


def test_photonics_error() -> None:
    # check error
    photo = core.Photonics("bragg", 16)
    np.testing.assert_raises(AssertionError, photo, np.zeros(12))
    with warnings.catch_warnings(record=True) as ws:
        output = photo(np.zeros(16))
        # one warning on Ubuntu, two warnings with Windows
        assert any(isinstance(w.message, RuntimeWarning) for w in ws)
    np.testing.assert_almost_equal(output, float("inf"))


@pytest.mark.parametrize("method", ["clipping", "arctan", "tanh", "constraint"])  # type: ignore
@pytest.mark.parametrize("name", ["bragg", "morpho", "chirped"])  # type: ignore
def test_no_warning(name: str, method: str) -> None:
    with warnings.catch_warnings(record=True) as w:
        core.Photonics(name, 24, bounding_method=method)
        assert not w, f"Got a warning at initialization: {w[0]}"


@testing.parametrized(
    # morpho=("morpho", 100, 1.1647),  # too slow
    chirped=("chirped", 150, 0.94439),
    bragg=("bragg", 2.5, 0.93216),
)
def test_photonics_values(name: str, value: float, expected: float) -> None:
    if name == "morpho" and os.environ.get("CIRCLECI", False):
        raise SkipTest("Too slow in CircleCI")
    photo = core.Photonics(name, 16)
    np.testing.assert_almost_equal(photo(value * np.ones(16)), expected, decimal=4)


GOOD_CHIRPED = [
    89.04887416,
    109.54188095,
    89.74520725,
    121.81700431,
    179.99830918,
    124.38222473,
    95.31017129,
    116.0239629,
    92.92345776,
    118.06108198,
    179.99965859,
    116.89288181,
    88.90191494,
    110.30816229,
    93.11974992,
    137.42629858,
    118.81810084,
    110.74139708,
    85.15270955,
    100.9382438,
    81.44070951,
    100.6382896,
    84.97336252,
    110.59252719,
    134.89164276,
    121.84205195,
    89.28450356,
    106.72776991,
    85.77168797,
    102.33562547,
]


EPS_AND_D = [  # photosic realistic
    2.0000,
    3.0000,
    2.1076,
    2.0000,
    3.0000,
    2.5783,
    2.0000,
    3.0000,
    2.0000,
    3.0000,
    90.0231,
    78.9789,
    72.8369,
    99.9577,
    82.7487,
    62.7583,
    104.1682,
    139.9002,
    93.3356,
    75.6039,
]


@testing.parametrized(
    morpho=("morpho", 1.127904, None),
    chirped=("chirped", 0.594587, None),
    good_chirped=("chirped", 0.275923, np.array([GOOD_CHIRPED])),  # supposed to be better
    bragg=("bragg", 0.96776, None),
    photosic_realistic=("cf_photosic_realistic", 0.0860257, np.array(EPS_AND_D).reshape((2, -1))),
    photosic_reference=("cf_photosic_reference", 0.431072, None),
)
def test_photonics_values_random(name: str, expected: float, data: tp.Optional[np.ndarray]) -> None:
    if name == "morpho" and os.environ.get("CIRCLECI", False):
        raise SkipTest("Too slow in CircleCI")
    size = data.size if data is not None else (16 if name != "morpho" else 4)
    photo = core.Photonics(name, size)
    if data is None:
        x = np.random.RandomState(12).normal(0, 1, size=size)
        candidate = photo.parametrization.spawn_child().set_standardized_data(x)
    else:
        candidate = photo.parametrization.spawn_child(new_value=data)
    np.testing.assert_almost_equal(photo(candidate.value), expected, decimal=4)
    np.testing.assert_almost_equal(photo.evaluation_function(candidate), expected, decimal=4)


def test_photosic_reference() -> None:
    debut = np.array([79, 102])
    fin = np.array([100, 70])
    dbr = np.tile(np.array([147, 120]), 3)
    X = np.concatenate([debut, dbr, fin])
    cf_test = photonics.cf_photosic_reference(X)
    np.testing.assert_almost_equal(cf_test, 0.09069397)


def test_photosic_realist() -> None:
    cf_test = photonics.cf_photosic_realistic(np.array(EPS_AND_D))
    np.testing.assert_almost_equal(cf_test, 0.08602574254532869)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pickle
import numpy as np
import pytest
import nevergrad as ng
import nevergrad.common.typing as tp
from nevergrad.functions import ArtificialFunction
from nevergrad.benchmark import experiments
from nevergrad.benchmark.xpbase import Experiment
from . import recaster
from . import optimizerlib


def fake_caller(func: tp.Callable[[int], int]) -> int:
    output = 0
    for k in range(10):
        output += func(k)
    return output


def test_automatic_thread_deletion() -> None:
    thread = recaster.MessagingThread(fake_caller)
    assert thread.is_alive()


def fake_cost_function(x: tp.ArrayLike) -> float:
    return float(np.sum(np.array(x) ** 2))


class FakeOptimizer(recaster.SequentialRecastOptimizer):
    def get_optimization_function(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.ArrayLike]:
        # create a new instance to avoid deadlock
        return self.__class__(self.parametrization, self.budget, self.num_workers)._optim_function

    def _optim_function(self, func: tp.Callable[..., tp.Any]) -> tp.ArrayLike:
        suboptim = optimizerlib.OnePlusOne(parametrization=2, budget=self.budget)
        recom = suboptim.minimize(func)
        return recom.get_standardized_data(reference=self.parametrization)


def test_recast_optimizer() -> None:
    optimizer = FakeOptimizer(parametrization=2, budget=100)
    optimizer.minimize(fake_cost_function)
    assert optimizer._messaging_thread is not None
    np.testing.assert_equal(optimizer._messaging_thread._thread.call_count, 100)


def test_recast_optimizer_with_error() -> None:
    optimizer = FakeOptimizer(parametrization=2, budget=100)
    np.testing.assert_raises(TypeError, optimizer.minimize)  # type: ignore


def test_recast_optimizer_and_stop() -> None:
    optimizer = FakeOptimizer(parametrization=2, budget=100)
    optimizer.ask()
    # thread is not finished... but should not hang!


def test_provide_recommendation() -> None:
    opt = optimizerlib.SQP(parametrization=2, budget=100)
    assert isinstance(
        opt.provide_recommendation(), ng.p.Parameter
    ), "Recommendation should be available from start"
    # the recommended solution should be the better one among the told points
    x1 = opt.ask()
    opt.tell(x1, 10)
    x2 = opt.ask()
    opt.tell(x2, 5)
    recommendation = opt.provide_recommendation()
    np.testing.assert_array_almost_equal(recommendation.value, x2.value)


def test_sqp_with_constraint() -> None:
    func = ArtificialFunction("ellipsoid", block_dimension=10, rotation=True, translation_factor=0.1)
    func.parametrization.register_cheap_constraint(experiments._Constraint("sum", as_bool=True))
    xp = Experiment(func, optimizer="ChainMetaModelSQP", budget=150, seed=4290846341)
    xp._run_with_error()


class ErroringSequentialRecastOptimizer(recaster.SequentialRecastOptimizer):
    def get_optimization_function(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.ArrayLike]:
        def optim_function(func: tp.Callable[..., tp.Any]) -> tp.ArrayLike:
            func(np.zeros(2))
            raise ValueError("ErroringOptimizer")

        return optim_function


class ErroringBatchRecastOptimizer(recaster.BatchRecastOptimizer):

    # pylint: disable=abstract-method

    def get_optimization_function(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.ArrayLike]:
        def optim_function(func: tp.Callable[..., tp.Any]) -> tp.ArrayLike:
            func(np.zeros((1, 2)))
            raise ValueError("ErroringOptimizer")

        return optim_function


def test_recast_optimizer_error() -> None:
    sequential = ErroringSequentialRecastOptimizer(parametrization=2, budget=100)
    sequential.tell(sequential.ask(), 4.2)
    with pytest.raises(ValueError, match="ErroringOptimizer"):
        sequential.ask()

    batch = ErroringBatchRecastOptimizer(parametrization=2, budget=100)
    batch.tell(batch.ask(), 4.2)
    with pytest.raises(ValueError, match="ErroringOptimizer"):
        batch.ask()


def _simple_multiobjective(x):
    return [np.sum(x**2), np.sum((x - 1) ** 2)]


@pytest.mark.parametrize("after_ask", [False, True])  # type: ignore
def test_recast_pickle(after_ask: bool) -> None:
    # Do 10 ask/tells and optionally another ask.
    optimizer = ng.optimizers.PymooNSGA2(parametrization=2, budget=300)
    optimizer.enable_pickling()
    optimizer.parametrization.random_state.seed(12)
    for _ in range(10):
        x = optimizer.ask()
        loss = _simple_multiobjective(*x.args, **x.kwargs)
        optimizer.tell(x, loss)
    if after_ask:
        x_active = optimizer.ask()
        x_active_2 = pickle.loads(pickle.dumps(x_active))

    # Copy the optimizer and copy the copy and complete both to a total
    # of 60 ask/tells
    optimizer_remade = pickle.loads(pickle.dumps(optimizer))
    optimizer_remade_remade = pickle.loads(pickle.dumps(optimizer_remade))
    for i in range(50):
        x = x_active if (i == 0 and after_ask) else optimizer_remade.ask()
        loss = _simple_multiobjective(*x.args, **x.kwargs)
        optimizer_remade.tell(x, loss)

    for i in range(50):
        x = x_active_2 if (i == 0 and after_ask) else optimizer_remade_remade.ask()
        loss = _simple_multiobjective(*x.args, **x.kwargs)
        optimizer_remade_remade.tell(x, loss)

    # Do a single optimization with 60 ask/tells
    optimizer_2 = ng.optimizers.PymooNSGA2(parametrization=2, budget=300)
    optimizer_2.parametrization.random_state.seed(12)
    for _ in range(60):
        x = optimizer_2.ask()
        loss = _simple_multiobjective(*x.args, **x.kwargs)
        optimizer_2.tell(x, loss)

    # Check the results all agree.
    assert optimizer_remade.num_ask == 60
    assert optimizer_remade.num_tell == 60
    assert optimizer_remade_remade.num_ask == 60
    assert optimizer_remade_remade.num_tell == 60
    assert optimizer_2.num_ask == 60
    assert optimizer_2.num_tell == 60

    pf1 = optimizer_remade.pareto_front()
    pf2 = optimizer_2.pareto_front()
    assert len(pf1) == len(pf2)
    for a_value, b_value in zip([i.value for i in pf1], [i.value for i in pf2]):
        assert np.allclose(a_value, b_value)

    pf11 = optimizer_remade_remade.pareto_front()
    pf2 = optimizer_2.pareto_front()
    assert len(pf11) == len(pf2)
    for a_value, b_value in zip([i.value for i in pf11], [i.value for i in pf2]):
        assert np.allclose(a_value, b_value)

    # Check that unpickling hasn't caused extra random draws.
    expected_rand = optimizer_2._rng.randint(999999)
    assert optimizer_remade._rng.randint(999999) == expected_rand
    assert optimizer_remade_remade._rng.randint(999999) == expected_rand

    with pytest.raises(ValueError, match="you should have asked"):
        pickle.dumps(optimizer_2)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import typing as tp
import networkx as nx
import cdt
from cdt.metrics import precision_recall, SHD
import nevergrad as ng
from ..base import ExperimentFunction


# pylint: disable=too-many-instance-attributes,too-many-arguments,too-many-locals
class CausalDiscovery(ExperimentFunction):
    """
    All parameters, except for the `generator` argument, are used to configure `acylicgraph`.
    For the setting of the acylic graph generator, please refer to https://fentechsolutions.github.io/CausalDiscoveryToolbox/html/data.html

    Parameters
    ----------
    generator: str
        'tuebingen', 'sachs', 'dream4' or 'acylicgraph'
    causal_mechanism: str
        'linear', 'polynomial', 'sigmoid_add', 'sigmoid_mix', 'gp_add', 'gp_mix', 'nn'
    noise: str or function
        Type of noise function to use in the generative process
    noise_coeff: float
        Proportion of noise
    npoints: int
        Total number of data points
    nodes: int
        Total number of nodes (i.e. variables)
    parents_max: int
        Maximum number of parent nodes
    expected_nedges: int
        Number of edge per node expected if erdos graph is used
    """

    def __init__(
        self,
        generator: str = "sachs",
        causal_mechanism: str = "linear",
        noise: str = "gaussian",
        noise_coeff: float = 0.4,
        npoints: int = 500,
        nodes: int = 20,
        parents_max: int = 5,
        expected_degree: int = 3,
    ) -> None:
        # Currently, there are three datasets as described in https://fentechsolutions.github.io/CausalDiscoveryToolbox/html/data.html
        _dataset = ["tuebingen", "sachs", "dream4"]
        assert generator in _dataset + ["acylicgraph"]
        assert causal_mechanism in [
            "linear",
            "polynomial",
            "sigmoid_add",
            "sigmoid_mix",
            "gp_add",
            "gp_mix",
            "nn",
        ]
        if generator in _dataset:
            self._data, self._ground_truth_graph = cdt.data.load_dataset(generator)
        else:
            _generator = cdt.data.AcyclicGraphGenerator(
                causal_mechanism=causal_mechanism,
                noise=noise,
                noise_coeff=noise_coeff,
                npoints=npoints,
                nodes=nodes,
                parents_max=parents_max,
                expected_degree=expected_degree,
            )
            self._data, self._ground_truth_graph = _generator.generate()

        self._nodes_list = list(self._ground_truth_graph.nodes())
        self._nvars = self._data.shape[1]
        param_links = ng.p.Choice([-1, 0, 1], repetitions=self._nvars * (self._nvars - 1) // 2)
        instru = ng.p.Instrumentation(network_links=param_links).set_name("")
        super().__init__(self.objective, instru)

    def objective(self, network_links: tp.Tuple[int]) -> float:
        output_graph = self.choices_to_graph(network_links)
        score = self.graph_score(output_graph)
        return -score

    def graph_score(self, test_graph: nx.DiGraph) -> float:
        pr_score, _ = precision_recall(self._ground_truth_graph, test_graph)
        shd_score = SHD(self._ground_truth_graph, test_graph)
        return float(pr_score - shd_score)  # Higher better

    def add_missing_nodes(self, graph):
        for n in set(self._nodes_list) - set(graph.nodes()):
            graph.add_node(n)
        return graph

    def choices_to_graph(self, network_links: tp.Tuple[int]) -> nx.DiGraph:
        output_graph = nx.DiGraph()
        k = 0
        for i in range(1, self._nvars):
            for j in range(i + 1, self._nvars):
                if network_links[k] == -1:
                    output_graph.add_edge(self._nodes_list[j], self._nodes_list[i])
                elif network_links[k] == +1:
                    output_graph.add_edge(self._nodes_list[i], self._nodes_list[j])
                k += 1
        return self.add_missing_nodes(output_graph)
